/* Snownews - A lightweight console RSS newsreader
 * 
 * Copyright 2003 Oliver Feiler <kiza@kcore.de>
 * http://kiza.kcore.de/software/snownews/
 *
 * interface.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <ncurses.h>

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "config.h"
#include "main.h"
#include "netio.h"
#include "interface.h"
#include "xmlparse.h"
#include "conversions.h"

#ifdef SUN
#	include "os-support.h"
#endif

extern char *browser;
extern struct keybindings keybindings;

/* Scrollable feed->description. */
struct scrolltext {
	char * line;
	struct scrolltext * next_ptr;
	struct scrolltext * prev_ptr;
};

/* Init the ncurses library. */
void InitCurses (void) {
	initscr();

	keypad (stdscr, TRUE);	/* Activate keypad 'cuz some howto told me to do so. */
	
	cbreak();				/* No buffering. */
	noecho();				/* Do not echo typed chars onto the screen. */
	clear();
	curs_set(0);			/* Hide cursor. */
	refresh();
}


/* Print text in statusbar. */
void UIStatus (char * text, int delay) {
	int i;
	
	attron (WA_REVERSE);
	/* Clear statusbar. */
	for (i = 0; i < COLS-1; i++) {
		mvaddch (LINES-1, i, ' ');
	}
	
	/* mvprintw (LINES-1, 1, text); */
	mvaddstr (LINES-1, 1, text);
	
	/* attroff is called here. If the calling function had it switched on,
	   switch it on again there! */
	attroff (WA_REVERSE);
	refresh();
	
	if (delay)
		sleep (delay);
}

char * UIOneLineEntryField (int x, int y) {
        char *text;

        text = malloc(512);
        
        /* UIStatus switches off attron! */
        attron (WA_REVERSE);
        echo();
        curs_set(1);

        move (y, x);
        /* Beware of hardcoded textlength size! */
        getnstr (text, 511);
        
        noecho();
        curs_set(0);
        attroff (WA_REVERSE);
        
        /* This memory needs to be freed in the calling function! */
        return text;
}

int UIChangeBrowser (void) {
	int centerx, centery, i, j, len;
	char *browserstring;
	
	/* malloc = 17 (strlen("Current setting: ") + browser
	   We will malloc a bigger junk, because other languages
	   might need longer strings and crash! */
	len = strlen(_("Current setting: %s")) + strlen(browser) + 1;
	browserstring = malloc (len);
	snprintf (browserstring, len, _("Current setting: %s"), browser);
	
	centerx = COLS / 2;
	centery = LINES / 2;

	/* Clear screen area we want to "draw" to. */
	attron (WA_REVERSE);
	for (i = 5; i <= 7; i++) {
		for (j = -(COLS/2-3); j <= (COLS/2-3); j++) {
				mvaddch (i, centerx+j, ' ');
		}	
	}
	
	UIStatus (browserstring, 0);
	free (browserstring);
	
	browserstring = UIOneLineEntryField (5, 6);
	
	if (strlen(browserstring) == 0) {
		free (browserstring);
		return 1;
	}
	
	browser = realloc (browser, strlen(browserstring)+1);
	strncpy (browser, browserstring, strlen(browserstring)+1);
	
	free (browserstring);
	
	return 0;
}


/* Popup window to add new RSS feed. */
int UIAddFeed (void) {
	char tmp[512];
	char *url;
	int centerx, centery;
	int i, j;
	struct feed *new_ptr;
	struct feed *tmp_ptr;

	centerx = COLS / 2;
	centery = LINES / 2;

	/* Clear screen area we want to "draw" to. */
	attron (WA_REVERSE);
	for (i = 5; i <= 7; i++) {
		for (j = -(COLS/2-3); j <= (COLS/2-3); j++) {
				mvaddch (i, centerx+j, ' ');
		}	
	}
	UIStatus (_("Enter URL of the feed you want to add. Blank line to abort."), 0);

	url = UIOneLineEntryField (5, 6);
	
	/* If read stringlength is ZARO (abort of action requested) return 2
	   and confuse the calling function. */
	if (strlen(url) == 0) {
		free (url);
		return 1;
	}
	
	/* Check of entered url is valid at all. */
	if (strstr(url, "http://") == NULL) {
		free (url);
		return 2;
	}

	new_ptr = malloc (sizeof(struct feed));
	/* getnstr does not return newline... says the docs. */
	new_ptr->feedurl = malloc (strlen(url)+1);
	new_ptr->feed = NULL;
	
	/* Set to NULL so xmlparse.c is happy when we try to feed there. */
	new_ptr->title = NULL;
	new_ptr->link = NULL;
	new_ptr->description = NULL;
	new_ptr->items = NULL;
	new_ptr->lastmodified = NULL;
			
	strncpy (new_ptr->feedurl, url, strlen(url)+1);
	new_ptr->next_ptr = first_ptr;
	if (first_ptr != NULL)
		first_ptr->prev_ptr = new_ptr;
	new_ptr->prev_ptr = NULL;
	first_ptr = new_ptr;
	
	/* Don't need url text anymore. */
	free (url);

	/* Download new feed and DeXMLize it. */	
	if ((UpdateFeed (new_ptr)) != 0) {
		if ((lasthttpstatus != 200) && (lasthttpstatus != 0)) {
			snprintf (tmp, sizeof(tmp), _("HTTP error %d - Could not update feed."), lasthttpstatus);
			UIStatus (tmp, 2);
		}
		tmp_ptr = first_ptr;
		if (first_ptr->next_ptr != NULL)
			 first_ptr->next_ptr->prev_ptr = NULL;
		first_ptr = first_ptr->next_ptr;
		free (tmp_ptr);
		return -1;
	}

	return 0;
}

void FeedInfo (struct feed * current_feed) {
	int centerx, i, j, len;
	char *hashme;			/* Hashed filename. */
	char *file;
	struct stat filetest;
	
	centerx = COLS / 2;
	
	hashme = malloc (strlen(current_feed->feedurl)+1);
	strncpy (hashme, current_feed->feedurl, strlen(current_feed->feedurl)+1);
	Hashify (hashme);
	
	len = (strlen(getenv("HOME")) + strlen(hashme) + 18);
	file = malloc (len);
	snprintf (file, len, "%s/.snownews/cache/%s", getenv("HOME"), hashme);

	/* Clear the help screen area and print the text on reverse background. */
	/*attron (WA_REVERSE);
	for (i = -4; i <= 5; i++) {
		for (j = -20; j <= 20; j++) {
				mvaddch (centery+i, centerx+j, ' ');
		}	
	}*/
	/* Clear screen area we want to "draw" to. */
	attron (WA_REVERSE);
	for (i = 0; i <= 7; i++) {
		for (j = -(COLS/2-5); j <= (COLS/2-5); j++) {
				mvaddch (4+i, centerx+j, ' ');
		}	
	}
	mvaddstr (5, centerx-((strlen(current_feed->title))/2), current_feed->title);
	mvaddstr (7, centerx-(COLS/2-7), current_feed->feedurl);
	if (current_feed->lastmodified != NULL)
		mvprintw (8, centerx-(COLS/2-7), _("Last updated: %s"), current_feed->lastmodified);	
	else
		mvaddstr (8, centerx-(COLS/2-7), _("No modification date."));
	
	if ((stat (file, &filetest)) != -1) {
		mvprintw (9, centerx-(COLS/2-7), _("In disk cache: %lld bytes"), (long long int) filetest.st_size);
	} else
		mvaddstr (9, centerx-(COLS/2-7), _("Not in disk cache."));
	
	UIStatus (_("Displaying feed information."), 0);
	
	free (file);
	free (hashme);

	getch();
}


/* View newsitem in scrollable window.
 * Speed of this code has been greatly increased in 1.2.1
 */
void UIDisplayItem (struct newsitem * current_item, struct feed * current_feed) {
	int i, j, len;
	int uiinput;
	char *syscall = NULL;			/* Needs to be set to NULL. */
	char *syscallnull = NULL;		/* gcc will issue "might be used uninitialized" otherwise. */
	char *escapetext = NULL;
	char *tmp;
	char *feeddescscreenlength;		/* Used to print a max of COLS-something
									   for feed->desc onto the screen. */
	char *newtext;
	char *newtextwrapped;
	char *textslice;
	int colums;
	struct scrolltext *textlist;
	struct scrolltext *cur_ptr;
	struct scrolltext *first_ptr = NULL;
	struct scrolltext *tmp_ptr;
	struct scrolltext *next_ptr;
	int ypos;
	int linenumber = 0;				/* First line on screen (scrolling). Ugly hack. */
	int maxlines = 0;
	char *freeme;					/* str pos pointer. */
	char *converted;				/* Converted from UTF-8 to local charset. */
	int updatelink;
	int rewrap;						/* Set to 1 if text needs to be rewrapped. */
	
	/* This flag tells whether we need to update with a new url string.
	   If it's 0 the code below will not be executed. */
	updatelink = 1;
	
	/* Activate rewrap. */
	rewrap = 1;
	
	while (1) {
		if (updatelink) {
			/* Need to do this inside while loop so it gets
			   updated if user uses <-,->,n,p keys. */
		   
			syscall = NULL;
			syscallnull = NULL;
			escapetext = NULL;
		   
			/* NULL pointer check! */
			if (current_item->link != NULL) {
				if ((strstr(current_item->link, "'")) == NULL) {
					len = strlen(current_item->link) + 3;
					escapetext = malloc (len);
					snprintf (escapetext, len, "\'%s\'", current_item->link);
					len = strlen(browser) + len + 1;
					syscall = malloc (len);
					snprintf (syscall, len, browser, escapetext);
				} else {
					len = strlen(browser) + 1;
					syscall = malloc (len);
					snprintf (syscall, len, browser, "");
				}
			} else {
				len = strlen(browser) + 1;
				syscall = malloc (len);
				snprintf (syscall, len, browser, "");
			}
			updatelink = 0;
		}
		
		ypos = 6;
		
		clear();
		attron (WA_REVERSE);
		for (i = 0; i < COLS; i++) {
			mvaddch (0, i, ' ');
		}
		mvprintw (0, 1, "* Snownews %s", VERSION);
		attroff (WA_REVERSE);

		/* If feed doesn't have a description print title instead. */
		attron (WA_BOLD);
		if (current_feed->description != NULL) {
			/* Define colums, because malloc three lines down might fail
			   and corrupt mem while resizing the window. */
			colums = COLS-10;
			/* This function only prints COLS-10 number of letters for each feed->desc. */
			if (strlen(current_feed->description) > colums) {
				feeddescscreenlength = malloc(colums+1);
				strncpy (feeddescscreenlength, current_feed->description, colums);
				feeddescscreenlength[colums] = '\0';
				converted = iconvert (feeddescscreenlength, "UTF-8", TARGET_CHARSET);
				if (converted == NULL)
					converted = strdup (feeddescscreenlength);
				mvprintw (2, 1, "%s...", converted);
				free (feeddescscreenlength);
				free (converted);
			} else {
				converted = iconvert (current_feed->description, "UTF-8", TARGET_CHARSET);
				if (converted == NULL)
					converted = strdup (current_feed->description);
				mvprintw (2, 1, converted);
				free (converted);
			}
		}
		else
			mvprintw (2, 1, current_feed->title);
		attroff (WA_BOLD);
		
		/* Don't print feed titles longer than screenwidth here. */
		colums = COLS-6;
		/* And don't crash if title is empty. */
		if (current_item->title != NULL) {
			if (strlen(current_item->title) > colums) {
				feeddescscreenlength = malloc(colums+1);
				strncpy (feeddescscreenlength, current_item->title, colums);
				feeddescscreenlength[colums] = '\0';
				converted = iconvert (feeddescscreenlength, "UTF-8", TARGET_CHARSET);
				if (converted == NULL)
					converted = strdup (feeddescscreenlength);
				mvprintw (4, 1, "%s...", converted);
				free (converted);
				free (feeddescscreenlength);
			} else {
				converted = iconvert (current_item->title, "UTF-8", TARGET_CHARSET);
				if (converted == NULL)
					converted = strdup (current_item->title);
				mvprintw (4, (COLS/2)-(strlen(converted)/2), "%s", converted);
				free (converted);
			}
		} else
			mvaddstr (4, 1, _("No title"));
		
		if (current_item->description == NULL)
			mvprintw (6, 1, _("No description available."));
		else {
			if (strlen(current_item->description) == 0)
				mvprintw (6, 1, _("No description available."));
			else {
				/* Only generate a new scroll list if we need to rewrap everything.
				   Otherwise just skip this block. */
				if (rewrap) {
					converted = iconvert (current_item->description, "UTF-8", TARGET_CHARSET);
					if (converted == NULL)
						converted = strdup (current_item->description);
					newtext = UIDejunk (converted);
					free (converted);
					newtextwrapped = WrapText (newtext, COLS-4);
					free (newtext);
					freeme = newtextwrapped;	/* Set ptr to str start so we can free later. */
					
					/* 
					 * Split newtextwrapped at \n and put into double linked list.
					 */
					while (1) {
						textslice = strsep (&newtextwrapped, "\n");
						
						/* Find out max number of lines text has. */
						maxlines++;
						
						if (textslice != NULL) {
							textlist = malloc (sizeof(struct scrolltext));
							
							textlist->line = malloc (strlen(textslice)+1);
							strcpy (textlist->line, textslice);
							
							/* Gen double list with new items at bottom. */
							textlist->next_ptr = NULL;
							if (first_ptr == NULL) {
								textlist->prev_ptr = NULL;
								first_ptr = textlist;
							} else {
								textlist->prev_ptr = first_ptr;
								while (textlist->prev_ptr->next_ptr != NULL)
									textlist->prev_ptr = textlist->prev_ptr->next_ptr;
								textlist->prev_ptr->next_ptr = textlist;
							}
						} else
							break;
					}
				
					free (freeme);
					rewrap = 0;
				}
				
				j = linenumber;
				/* We sould now have the linked list setup'ed... hopefully. */
				for (cur_ptr = first_ptr; cur_ptr != NULL; cur_ptr = cur_ptr->next_ptr) {
					/* Skip lines we don't need to display. */
					if (j > 0) {
						j--;
						continue;
					}
					if (ypos > LINES-4)
						continue;
					mvaddstr (ypos, 2, cur_ptr->line);
					ypos++;
				}
			}
		}
		

		attron (WA_BOLD);
		mvprintw (LINES-2, 1, "-> %s", syscall);
		attroff (WA_BOLD);
		
		/* Set current item to read. */
		current_item->readstatus = 1;
		
		UIStatus (_("Press 'q' or Enter to return to previous screen. Hit 'h' for help screen."), 0);
		uiinput = getch();
		
		if (uiinput == 'h')
			UIDisplayItemHelp();
		if ((uiinput == '\n') ||
			(uiinput == keybindings.quit)) {
			free (escapetext);
			free (syscall);
			return;
		}
		if (uiinput == keybindings.urljump) {
			len = strlen(syscall) + strlen(_("Executing %s")) + 1;
			tmp = malloc (len);
			snprintf (tmp, len, _("Executing %s"), syscall);
			UIStatus (tmp, 0);
			len = len + 13;
			syscallnull = malloc (len);
			snprintf (syscallnull, len, "%s 2>/dev/null", syscall);
			system(syscallnull);
			InitCurses();
			free (tmp);
			free (syscallnull);
		}
		if ((uiinput == keybindings.next) ||
			(uiinput == KEY_RIGHT)) {
			if (current_item->next_ptr != NULL) {
				current_item = current_item->next_ptr;
				linenumber = 0;
				/* Set flag to update with new url string. */
				updatelink = 1;
				rewrap = 1;
				maxlines = 0;
			} else {
				/* Setting rewrap to 1 to get the free block below executed. */
				rewrap = 1;
				uiinput = ungetch(keybindings.quit);
			}
		}
		if ((uiinput == keybindings.prev) ||
			(uiinput == KEY_LEFT)) {
			if (current_item->prev_ptr != NULL) {
				current_item = current_item->prev_ptr;
				linenumber = 0;
				/* Set flag to update with new url string. */
				updatelink = 1;
				rewrap = 1;
				maxlines = 0;
			} else {
				/* Setting rewrap to 1 to get the free block below executed. */
				rewrap = 1;
				uiinput = ungetch(keybindings.quit);
			}
		}
		/* Scroll by one page. */
		if (uiinput == KEY_NPAGE) {
			for (i = 0; i < LINES-9; i++) {
				if (linenumber+(LINES-7) < maxlines)
					linenumber++;
			}
		}
		if (uiinput == KEY_PPAGE) {
			for (i = 0; i < LINES-9; i++) {
				if (linenumber >= 1)
					linenumber--;
			}
		}
		if (uiinput == 'A')
			UIAbout();
		if (uiinput == KEY_UP) {
			if (linenumber >= 1)
				linenumber--;
		}
		if (uiinput == KEY_DOWN) {
			if (linenumber+(LINES-7) < maxlines)
				linenumber++;
		}
		if (uiinput == KEY_RESIZE) {
			rewrap = 1;
			/* Reset maxlines, otherwise the program will scroll to far down. */
			maxlines = 0;
		}
		
		/* Free the linked list structure if we need to rewrap the text block. */
		if (rewrap) {
			/* Free textlist list before we start loop again. */
			for (cur_ptr = first_ptr; cur_ptr != NULL; cur_ptr = next_ptr) {
				tmp_ptr = cur_ptr;
				if (cur_ptr->next_ptr != NULL)
					first_ptr = cur_ptr->next_ptr;
				free (tmp_ptr->line);
				next_ptr = cur_ptr->next_ptr;
				free (tmp_ptr);
			}
			/* first_ptr must be set to NULL again!
			   = back to init value. */
			first_ptr = NULL;
		}
		/* Only execute free if updatelink == 1 */
		if (updatelink) {
			/* Only free escapetext if we used it above! */
			if (escapetext != NULL)
				free (escapetext);
			free (syscall);
		}
	}
}


void UIDisplayFeed (struct feed * current_feed) {
	char *tmp;
	char *syscall, *syscallnull, *escapetext;
	int uiinput = 0;
	int i, ypos, len;
	struct newsitem *curitem_ptr;
	struct newsitem *highlighted;
	struct newsitem *tmp_highlighted;
	struct newsitem *tmp_first;
	struct newsitem *first_scr_ptr;
	struct newsitem *markasread;
	int itemnum;
	int highlightnum = 1;
	char *feeddescscreenlength;		/* Used to print a max of COLS-something
									   for feed->desc onto the screen. */
	char *itemcut;					/* Don't print items longer than COLS. */
	int colums;
	char *converted;				/* Converted from UTF-8 to local charset. */
	
	syscall = NULL;
	syscallnull = NULL;
	escapetext = NULL;
	
	/* We don't need to check for NULL pointer since
	   current_feed->link will never be NULL. */
	if ((strstr(current_feed->link, "'")) == NULL) {
		len = strlen(current_feed->link) + 3;
		escapetext = malloc (len);
		snprintf (escapetext, len, "\'%s\'", current_feed->link);
		len = strlen(browser) + len + 1;
		syscall = malloc (len);
		snprintf (syscall, len, browser, escapetext);
	} else {
		len = strlen(browser) + 1;
		syscall = malloc (len);
		snprintf (syscall, len, browser, "");
	}
	
	/* Set highlighted to first_ptr at the beginning.
	   Otherwise bad things (tm) will happen. */
	first_scr_ptr = current_feed->items;
	highlighted = current_feed->items;
	
	/* Select first unread item if we enter feed view or
	   leave first item active if there is no unread. */
	tmp_highlighted = highlighted;
	tmp_first = first_scr_ptr;
	highlighted = first_scr_ptr;
	/* Moves highlight to next unread item. */
	
	/* Check if we have no items at all! */
	if (highlighted != NULL) {
		while (highlighted->next_ptr != NULL) {
			if (highlighted->readstatus == 1) {
				highlighted = highlighted->next_ptr;
				
				highlightnum++;
				if (highlightnum+1 > LINES-7) {
					/* Fell off screen. */	
					if (first_scr_ptr->next_ptr != NULL)
						first_scr_ptr = first_scr_ptr->next_ptr;
				}	
				
			} else
				break;				
		}
	}
	/* If *highlighted points to a read entry now we have hit the last
	   entry and there are no unread items anymore. Restore the original
	   location in this case. */
	/* Check if we have no items at all! */
	if (highlighted != NULL) {
		if (highlighted->readstatus == 1) {
			highlighted = tmp_highlighted;
			first_scr_ptr = tmp_first;
		}
	}
	
	while (1) {
		clear();
		
		/* This clears/inverts the first line and prints the app name/version. */
		attron (WA_REVERSE);
		for (i = 0; i < COLS; i++) {
			mvaddch (0, i, ' ');
		}
		mvprintw (0, 1, "* Snownews %s", VERSION);
		attroff (WA_REVERSE);
		
		attron (WA_BOLD);
		if (current_feed->description != NULL) {
			/* Print a max of COLS-something chars.
			   This function might cause overflow if user resizes screen
			   during malloc and strncpy. To prevent this we use a private
			   colums variable. */
			colums = COLS-10;
			if (strlen(current_feed->description) > colums) {
				feeddescscreenlength = malloc(colums+1);
				strncpy (feeddescscreenlength, current_feed->description, colums);
				feeddescscreenlength[colums] = '\0';
				converted = iconvert (feeddescscreenlength, "UTF-8", TARGET_CHARSET);
				/* Fallback: duplicate string into converted so free won't crash.
				   View the unconverted item instead if iconvert fails. */
				if (converted == NULL)
					converted = strdup (feeddescscreenlength);
				mvprintw (2, 1, "%s...", converted);
				free (feeddescscreenlength);
				free (converted);
			} else {
				converted = iconvert (current_feed->description, "UTF-8", TARGET_CHARSET);
				if (converted == NULL)
					converted = strdup (current_feed->description);
				mvprintw (2, 1, "%s", converted);
				free (converted);
			}
		} else {
			mvprintw (2, 1,	"%s", current_feed->title);
		}
		attroff (WA_BOLD);
		
		/* We start the item list at line 5. */
		ypos = 4;
		itemnum = 1;
		
		colums = COLS-6;			/* Cut max item length. */
		/* Print unread entries in bold. */
		for (curitem_ptr = first_scr_ptr; curitem_ptr != NULL; curitem_ptr = curitem_ptr->next_ptr) {
			if (curitem_ptr->readstatus != 1)
				attron (WA_BOLD);
			if (curitem_ptr == highlighted) {
				highlightnum = itemnum;
				attron (WA_REVERSE);
				for (i = 0; i < COLS-1; i++)
					mvaddch (ypos, i, ' ');
			}
			
			/* Check for empty <title> */
			if (curitem_ptr->title != NULL) {
				if (strlen(curitem_ptr->title) > colums) {
					itemcut = malloc(colums+1);
					strncpy (itemcut, curitem_ptr->title, colums);
					itemcut[colums] = '\0';
					converted = iconvert (itemcut, "UTF-8", TARGET_CHARSET);
					if (converted == NULL)
						converted = strdup (itemcut);
					mvprintw (ypos, 1, "%s...", converted);
					free (converted);
					free (itemcut);
				} else {
					converted = iconvert (curitem_ptr->title, "UTF-8", TARGET_CHARSET);
					if (converted == NULL)
						converted = strdup (curitem_ptr->title);
					mvaddstr (ypos, 1, converted);
					free (converted);
				}
			} else
				mvaddstr (ypos, 1, _("No title"));
			ypos++;
			if (curitem_ptr == highlighted)
				attroff (WA_REVERSE);
			if (curitem_ptr->readstatus != 1)
				attroff (WA_BOLD);
				
			if (itemnum >= LINES-7)
				break;
			itemnum++;
		}
		refresh();
		
		attron (WA_BOLD);
		mvprintw (LINES-2, 1, "-> %s", syscall);
		attroff (WA_BOLD);
		
		UIStatus (_("Press 'q' to return to main menu, 'h' to show help."), 0);
		
		uiinput = getch();
		
		if (uiinput == 'h')
			UIDisplayFeedHelp();
		if (uiinput == keybindings.quit) {
			/* Only free these when we return!
			   Otherwise this would introduce multifrees. 8) */
			free (syscall);
			free (escapetext);
			return;
		}
		if ((uiinput == KEY_UP) ||
			(uiinput == keybindings.prev)) {
			/* Check if we have no items at all! */
			if (highlighted != NULL) {
				if (highlighted->prev_ptr != NULL) {
					highlighted = highlighted->prev_ptr;
					if (highlightnum-1 < 1) {
						/* First scr entry. */
						if (first_scr_ptr->prev_ptr != NULL)
							first_scr_ptr = first_scr_ptr->prev_ptr;
						}
				}
			}
		}
		if ((uiinput == KEY_DOWN) ||
			(uiinput == keybindings.next)) {
			/* Check if we have no items at all! */
			if (highlighted != NULL) {
				if (highlighted->next_ptr != NULL) {
					highlighted = highlighted->next_ptr;
					if (highlightnum+1 > LINES-7) {
						/* Fell off screen. */	
						if (first_scr_ptr->next_ptr != NULL)
							first_scr_ptr = first_scr_ptr->next_ptr;
					}	
				}
			}
		}
		/* Move highlight one page up/down == LINES-9 */
		if (uiinput == KEY_NPAGE) {
			if (highlighted != NULL) {
				for (i = highlightnum; i <= LINES-8; i++) {
					if (highlighted->next_ptr != NULL) {
						highlighted = highlighted->next_ptr;
					} else
						break;
				}
			}
			first_scr_ptr = highlighted;
		}
		if (uiinput == KEY_PPAGE) {
			if (highlighted != NULL) {
				for (i = highlightnum; i <= LINES-8; i++) {
					if (highlighted->prev_ptr != NULL) {
						highlighted = highlighted->prev_ptr;
					} else
						break;
				}
			}
			first_scr_ptr = highlighted;
		}
		if (uiinput == '\n') {
			/* Check if we have no items at all!
			   Don't even try to view a non existant item. */
			if (highlighted != NULL) {
				UIDisplayItem (highlighted, current_feed);
				tmp_highlighted = highlighted;
				tmp_first = first_scr_ptr;
				
				highlighted = first_scr_ptr;
					
				/* Moves highlight to next unread item. */
				while (highlighted->next_ptr != NULL) {
					if (highlighted->readstatus == 1) {
						highlighted = highlighted->next_ptr;
						
						highlightnum++;
						if (highlightnum+1 > LINES-7) {
							/* Fell off screen. */	
							if (first_scr_ptr->next_ptr != NULL)
								first_scr_ptr = first_scr_ptr->next_ptr;
						}	
						
					} else
						break;				
				}
				
				/* If *highlighted points to a read entry now we have hit the last
				   entry and there are no unread items anymore. Restore the original
				   location in this case. */
				if (highlighted->readstatus == 1) {
					highlighted = tmp_highlighted;
					first_scr_ptr = tmp_first;
				}
			}
		}
		if (uiinput == keybindings.reload) {
			UpdateFeed (current_feed);
			highlighted = current_feed->items;
			/* Reset first_scr_ptr if reloading. */
			first_scr_ptr = current_feed->items;
		}
		if (uiinput == keybindings.urljump) {
			/* Copy "Executing lynx URL" into tmp and show it,
			   copy "lynx URL 2>/dev/null" into syscallnull and execute via system. */
			len = strlen(syscall) + strlen(_("Executing %s")) + 1;
			tmp = malloc (len);
			snprintf (tmp, len, _("Executing %s"), syscall);
			UIStatus (tmp, 0);
			len = len + 13;
			syscallnull = malloc (len);
			snprintf (syscallnull, len, "%s 2>/dev/null", syscall);
			system(syscallnull);
			InitCurses();
			free (tmp);
			free (syscallnull);
		}
		if (uiinput == keybindings.markread) {
			/* Mark everything read. */
			for (markasread = current_feed->items; markasread != NULL; markasread = markasread->next_ptr) {
				markasread->readstatus = 1;
			}
		}		
		if (uiinput == 'A')
			UIAbout();
		if (uiinput == keybindings.feedinfo)
			FeedInfo(current_feed);
		
		/* DON NOT free escapetext and syscall here. It would result in multiple frees! */
	}	
}

int UIDeleteFeed (char * feedname) {
	int i, j;

	/* Clear screen area we want to "draw" to. */
	attron (WA_REVERSE);
	for (i = 5; i <= 8; i++) {
		for (j = -(COLS/2-3); j <= (COLS/2-3); j++) {
				mvaddch (i, (COLS/2)+j, ' ');
		}	
	}
	mvaddstr (6, COLS/2-21, _("Are you sure you want to delete this feed?"));
	mvprintw (7, 5, "%s", feedname);
	
	UIStatus (_("Type 'y' to delete, any other key to abort."), 0);
	
	if (getch() == 'y')
		return 1;
	else
		return 0;
}


void UIMainInterface (void) {
	int i, ypos, len;
	int newcount;		/* Number of new articles. */
	int uiinput = 0;
	char tmp[1024];
	char *file, *hashme;
	struct feed *cur_ptr;
	struct newsitem *curitem_ptr;
	struct newsitem *markasread;
	struct feed *highlighted;
	struct feed *first_scr_ptr;		/* First pointer on top of screen.
									   Used for scrolling. */
	struct feed *removed;			/* Removed struct tmp pointer. */
	int itemnum;					/* Num of items on screen. Max is LINES-4 */
	int highlightnum = 1;
	int columns;
	char *maxlinesize;

	first_scr_ptr = first_ptr;
	highlighted = first_ptr;
		
	while (1) {
		clear();				/* Remove remaining junk from screen. */
		
		attron (WA_REVERSE);	
		for (i = 0; i < COLS; i++) {
			mvaddch (0, i, ' ');
		}
		mvprintw (0, 1, "* Snownews %s", VERSION);
		attroff (WA_REVERSE);
	
		ypos = 2;
		itemnum = 1;
	
		for (cur_ptr = first_scr_ptr; cur_ptr != NULL; cur_ptr = cur_ptr->next_ptr) {
			if (cur_ptr == highlighted) {
				highlightnum = itemnum;
				attron (WA_REVERSE);
				for (i = 0; i < COLS-1; i++)
					mvaddch (ypos, i, ' ');
			}
			
			columns = COLS-9-strlen(_("new"));
			if (strlen(cur_ptr->title) > columns) {
				maxlinesize = malloc(columns+1);
				strncpy (maxlinesize, cur_ptr->title, columns);
				maxlinesize[columns] = '\0';
				mvprintw (ypos, 1, "%s...", maxlinesize);
				free (maxlinesize);
			} else
				mvaddstr (ypos, 1, cur_ptr->title);
			
			newcount = 0;
			for (curitem_ptr = cur_ptr->items; curitem_ptr != NULL; curitem_ptr = curitem_ptr->next_ptr) {
				if (curitem_ptr->readstatus == 0)
					newcount++;
			}
			/* Execute this _after_ the for loop. Otherwise it'll suck CPU like crazy! */
			if (newcount != 0) {
				snprintf (tmp, sizeof(tmp), _("%3d new"), newcount);
				mvaddstr (ypos, COLS-5-strlen(_("new")), tmp);
			}
			ypos++;
			if (cur_ptr == highlighted)
				attroff (WA_REVERSE);
			
			if (itemnum >= LINES-4)
				break;
			itemnum++;
		}
		refresh();

		UIStatus (_("Press 'h' for help window."), 0);
		
		uiinput = getch();
		
		if (uiinput == keybindings.quit)
			MainQuit (NULL, NULL);
		if (uiinput == keybindings.reload) {
			/* Refresh highlighted feed. */
			if (UpdateFeed (highlighted) != 0) {
				if (lasthttpstatus == 0) {
					UIStatus (_("Aborted."), 1);
					continue;
				}
				if (lasthttpstatus == 304)
					continue;
				if (lasthttpstatus != 200) {
					snprintf (tmp, sizeof(tmp), _("HTTP error %d - Could not update feed."), lasthttpstatus);
					UIStatus (tmp, 2);
					continue;
				}
				if (lasthttpstatus != 0) {
					snprintf (tmp, sizeof(tmp), _("Could not update feed %s - %s"), highlighted->title, strerror(errno));
					UIStatus (tmp, 2);
				}
			}
		}
		if (uiinput == keybindings.reloadall)
			UpdateAllFeeds();
		if (uiinput == keybindings.addfeed) {
			switch (UIAddFeed()) {
				case 0:
					UIStatus (_("Successfully added new item..."), 1);
					
					/* If this was the first element added set highlighted to
			   		this one. */
			   		/* Why was this if == NULL added? */
					/* if (highlighted == NULL) { */
						highlighted = first_ptr;
					/* } */
					first_scr_ptr = first_ptr;
					
					break;
				case 1:
					UIStatus (_("Aborted."), 1);
					break;
				case 2:
					UIStatus (_("Invalid URL! Please add http:// if you forgot this."), 2);
					break;
				case -1:
				default:
					break;
			}
		}
		if (uiinput == 'h')
			UIHelpScreen();
		if (uiinput == keybindings.deletefeed) {
			if (UIDeleteFeed(highlighted->title) == 1) {
				/* Do it! */
				removed = highlighted;
				
				/* Remove cachefile from filesystem. */
				
				hashme = malloc (strlen(highlighted->feedurl)+1);
				strncpy (hashme, highlighted->feedurl, strlen(highlighted->feedurl)+1);
				Hashify (hashme);
	
				len = (strlen(getenv("HOME")) + strlen(hashme) + 18);
				file = malloc (len);
				snprintf (file, len, "%s/.snownews/cache/%s", getenv("HOME"), hashme);
				unlink (file);
				
				/* Unlink pointer from chain. */
				if (highlighted == first_ptr) {
					/* first element */
					if (first_ptr->next_ptr != NULL) {
						first_ptr = first_ptr->next_ptr;
						first_scr_ptr = first_ptr;
						highlighted->next_ptr->prev_ptr = NULL;
					} else {
						first_ptr = NULL;
						first_scr_ptr = NULL;
					}
				} else if (highlighted->next_ptr == NULL) {
					/* last element */
					highlighted->prev_ptr->next_ptr = NULL;
				} else {
					/* element inside list */
					highlighted->prev_ptr->next_ptr = highlighted->next_ptr;
					highlighted->next_ptr->prev_ptr = highlighted->prev_ptr;
				}
				highlighted = first_ptr;
				
				/* free (removed) pointer */
				while (removed->items->next_ptr != NULL) {
					removed->items = removed->items->next_ptr;
					free (removed->items->prev_ptr->title);
					free (removed->items->prev_ptr->link);
					free (removed->items->prev_ptr->description);
					free (removed->items->prev_ptr);
				}
				free (removed->feedurl);
				free (removed->feed);
				free (removed->title);
				free (removed->link);
				free (removed->description);
				free (removed);
				
			} else {
				UIStatus (_("Aborted."), 1);
			}
		}
		if (uiinput == '\n') {
			/* Select this feed, open and view entries. */
			/* Update disk cache if needed. */
			if (highlighted != NULL)
				UIDisplayFeed (highlighted);
		}
		if ((uiinput == KEY_UP) ||
			(uiinput == keybindings.prev)) {
			if (highlighted != NULL) {
				if (highlighted->prev_ptr != NULL) {
					highlighted = highlighted->prev_ptr;
					if (highlightnum-1 < 1) {
						/* Reached first onscreen entry. */
						if (first_scr_ptr->prev_ptr != NULL)
							first_scr_ptr = first_scr_ptr->prev_ptr;
					}
				}
			}
		}
		if ((uiinput == KEY_DOWN) ||
			(uiinput == keybindings.next)) {
			if (highlighted != NULL) {
				if (highlighted->next_ptr != NULL) {
					highlighted = highlighted->next_ptr;
					if (highlightnum+1 > LINES-4) {
						/* If we fall off the screen, advance first_scr_ptr
						   to next entry. */
						if (first_scr_ptr->next_ptr != NULL)
							first_scr_ptr = first_scr_ptr->next_ptr;
					}
				}
			}
		}
		/* Move highlight one page up/down == LINES-6 */
		if (uiinput == KEY_NPAGE) {
			if (highlighted != NULL) {
				for (i = highlightnum; i <= LINES-5; i++) {
					if (highlighted->next_ptr != NULL) {
						highlighted = highlighted->next_ptr;
					} else 
						break;
				}
			}
			first_scr_ptr = highlighted;
		}
		if (uiinput == KEY_PPAGE) {
			if (highlighted != NULL) {
				for (i = highlightnum; i <= LINES-5; i++) {
					if (highlighted->prev_ptr != NULL) {
						highlighted = highlighted->prev_ptr;
					} else
						break;
				}
			}
			first_scr_ptr = highlighted;
		}
		if (uiinput == keybindings.moveup) {
			/* Move item up. */
			if (highlighted->prev_ptr != NULL) {
				SwapPointers (highlighted, highlighted->prev_ptr);
				highlighted = highlighted->prev_ptr;
					if (highlightnum-1 < 1) {
						if (first_scr_ptr->prev_ptr != NULL)
							first_scr_ptr = first_scr_ptr->prev_ptr;
					}
			}
		}
		if (uiinput == keybindings.movedown) {
			/* Move item down. */
			if (highlighted->next_ptr != NULL) {
				SwapPointers (highlighted, highlighted->next_ptr);
				highlighted = highlighted->next_ptr;
				if (highlightnum+1 > LINES-4) {
					if (first_scr_ptr->next_ptr != NULL)
							first_scr_ptr = first_scr_ptr->next_ptr;
				}
			}
		}
		if (uiinput == keybindings.dfltbrowser) {
			if (UIChangeBrowser()) {
				UIStatus (_("Aborted."), 1);
			}
		}
		if (uiinput == keybindings.markread) {
			for (cur_ptr = first_scr_ptr; cur_ptr != NULL; cur_ptr = cur_ptr->next_ptr) {
				for (markasread = cur_ptr->items; markasread != NULL; markasread = markasread->next_ptr) {
						markasread->readstatus = 1;
				}
			}
		}
		if (uiinput == 'A')
			UIAbout();

	}
}

void UIHelpScreen (void) {
	int centerx, centery;				/* Screen center x/y coordinate. */
	int i, j;
	int userinput;
	
	centerx = COLS / 2;
	centery = LINES / 2;

	/* Clear the help screen area and print the text on reverse background. */
	attron (WA_REVERSE);
	for (i = -4; i <= 5; i++) {
		for (j = -20; j <= 20; j++) {
				mvaddch (centery+i, centerx+j, ' ');
		}	
	}
	mvprintw (centery-3, centerx-18, _("%c:    Add RSS feed..."), keybindings.addfeed);	
	mvprintw (centery-2, centerx-18, _("%c:    Delete highlighted RSS feed..."), keybindings.deletefeed);
	mvprintw (centery-1, centerx-18, _("%c:    Reload all feeds"), keybindings.reloadall);
	mvprintw (centery,   centerx-18, _("%c:    Reload this feed"), keybindings.reload);
	mvprintw (centery+1, centerx-18, _("%c:    Mark all read"), keybindings.markread);
	mvprintw (centery+2, centerx-18, _("%c:    Change default browser..."), keybindings.dfltbrowser);
	mvprintw (centery+3, centerx-18, _("%c, %c: Move item up, down"), keybindings.moveup, keybindings.movedown);
	mvprintw (centery+4, centerx-18, _("%c:    Quit program"), keybindings.quit);
	attroff (WA_REVERSE);

	UIStatus (_("Press the any(tm) key to exit help screen."), 0);
	userinput = getch();
	
	/* Return input back into input queue so it gets automatically
	   executed. */
	if ((userinput != '\n') && (userinput != 'h') && (userinput != 'q'))
		ungetch(userinput);
}

void UIDisplayFeedHelp (void) {
	int centerx, centery;				/* Screen center x/y coordinate. */
	int i, j;
	int userinput;
	
	centerx = COLS / 2;
	centery = LINES / 2;

	/* Clear the help screen area and print the text on reverse background. */
	attron (WA_REVERSE);
	for (i = -4; i <= 5; i++) {
		for (j = -17; j <= 17; j++) {
				mvaddch (centery+i, centerx+j, ' ');
		}	
	}
	mvprintw (centery-3, centerx-15, _("%c, up:    Previous item"), keybindings.prev);	
	mvprintw (centery-2, centerx-15, _("%c, down:  Next item"), keybindings.next);
	mvaddstr (centery-1, centerx-15, _("enter:    View item"));
	mvprintw (centery,   centerx-15, _("%c:        Reload this feed"), keybindings.reload);
	mvprintw (centery+1, centerx-15, _("%c:        Open homepage"), keybindings.urljump);
	mvprintw (centery+2, centerx-15, _("%c:        Mark all read"), keybindings.markread);
	mvprintw (centery+3, centerx-15, _("%c:        Show feed info..."), keybindings.feedinfo);
	mvprintw (centery+4, centerx-15, _("%c:        Return to main menu"), keybindings.quit);
	attroff (WA_REVERSE);

	UIStatus (_("Press the any(tm) key to exit help screen."), 0);
	userinput = getch();
	if ((userinput != '\n') && (userinput != 'h'))
		ungetch(userinput);
}

void UIDisplayItemHelp (void) {
	int centerx, centery;				/* Screen center x/y coordinate. */
	int i, j;
	int userinput;
	
	centerx = COLS / 2;
	centery = LINES / 2;

	/* Clear the help screen area and print the text on reverse background. */
	attron (WA_REVERSE);
	for (i = -3; i <= 3; i++) {
		for (j = -17; j <= 17; j++) {
				mvaddch (centery+i, centerx+j, ' ');
		}	
	}
	mvaddstr (centery-2, centerx-15, _("up, down: Scroll text"));
	mvprintw (centery-1, centerx-15, _("%c, left:  Previous item"), keybindings.prev);	
	mvprintw (centery,   centerx-15, _("%c, right: Next item"), keybindings.next);
	mvprintw (centery+1, centerx-15, _("%c:        Open link..."), keybindings.urljump);
	mvprintw (centery+2, centerx-15, _("%c, enter: Return to overview"), keybindings.quit);
	attroff (WA_REVERSE);

	UIStatus (_("Press the any(tm) key to exit help screen."), 0);
	userinput = getch();
	if ((userinput != '\n') && (userinput != 'h'))
		ungetch(userinput);
}


void UIAbout (void) {
	int ypos, xpos;
	
	clear();				/* Get the crap off the screen to make room
							   for our wonderful ASCII logo. :) */

	ypos = 2;
	xpos = COLS/2 - 40;
	
	if (COLS < 80) {
		mvprintw (0, 0, _("Need at least 80 COLS terminal, sorry!"));
		refresh();
		getch();
		return;
	}
	
	/* 90 COLS logo
	mvaddstr (ypos,   xpos, "  _________ ________     ______ ___________ ________     _______ ___________   _________");
	mvaddstr (ypos+1, xpos, " /  ______/ \\       \\   /      \\\\  |     | \\\\       \\   /   ___/ \\  |     | \\ /  ______/");
	mvaddstr (ypos+2, xpos, " \\_____  \\   \\   |   \\ /   /\\   \\\\    |    / \\   |   \\ /      /   \\    |    / \\_____  \\");
	mvaddstr (ypos+3, xpos, " /        \\  /   |    \\\\   \\/   / \\   |   /  /   |    \\\\   ___\\    \\   |   /  /        \\");
	mvaddstr (ypos+4, xpos, "/_______  / / ___|____/ \\______/  /___|__/  / ___|____/ \\_____ \\   /___|__/  /_______  /");
	mvaddstr (ypos+5, xpos, "        \\/  \\/                              \\/                \\/                     \\/");
	*/
	
	/* 80 COLS logo */
	mvaddstr (ypos,   xpos, "  ________ _______     ____ __________ _______     ______ __________   ________");
	mvaddstr (ypos+1, xpos, " /  _____/ \\      \\   /    \\\\  |    | \\\\      \\   /  ___/ \\  |    | \\ / ______/");
	mvaddstr (ypos+2, xpos, " \\____  \\   \\   |  \\ /  /\\  \\\\   |    / \\   |  \\ /     /   \\   |    / \\____  \\");
	mvaddstr (ypos+3, xpos, " /       \\  /   |   \\\\  \\/  / \\  |   /  /   |   \\\\  ___\\    \\  |   /  /       \\");
	mvaddstr (ypos+4, xpos, "/______  / / ___|___/ \\____/  /__|__/  / ___|___/ \\____ \\   /__|__/  /______  /");
	mvaddstr (ypos+5, xpos, "       \\/  \\/                          \\/              \\/                   \\/");
	mvprintw (ypos+7, COLS/2-(strlen("Version")+strlen(VERSION)+1)/2, "Version %s", VERSION);
	
	mvaddstr (ypos+10, COLS/2-(strlen(_("Brought to you by:")))/2, _("Brought to you by:"));
	mvaddstr (ypos+12, COLS/2-(strlen(_("Main code")))/2, _("Main code"));
	mvaddstr (ypos+13, COLS/2-6, "Oliver Feiler");
	mvaddstr (ypos+15, COLS/2-(strlen(_("Additional code")))/2, _("Additional code"));
	mvaddstr (ypos+16, COLS/2-4, "Rene Puls");
	mvaddstr (ypos+18, COLS/2-(strlen(_("Translation team")))/2, _("Translation team"));
	mvaddstr (ypos+19, COLS/2-31, "Oliver Feiler, Frank van der Loo, Pascal Varet, Simon Isakovic");

	getch();
}
